from typing import Any, Callable, Optional

import rtlsdr
from numpy.typing import NDArray

from skao_trt.gain_search import find_optimal_gain
from skao_trt.record_data import record_power_spectrum
from skao_trt.sdr_wrapper import MockRtlSdr


class ObsManager:
    """
    Manages the recording, storing and plotting the data.
    Does not perform any exception handling, this is deferred to the caller.
    """

    def __init__(
        self,
        mock_device: bool = False,
    ) -> None:
        self.sdr_class = MockRtlSdr if mock_device else rtlsdr.RtlSdr
        self.gain = None
        self.baseline_power_spectrum = None

    @property
    def ready_to_record(self) -> bool:
        return self.gain is not None

    @property
    def ready_to_observe(self) -> bool:
        return self.baseline_power_spectrum is not None

    def find_optimal_gain(
        self,
        callback: Optional[Callable[[float], Any]] = None,
    ) -> float:
        self.gain = find_optimal_gain(self.sdr_class, callback=callback)

    def record_baseline_power_spectrum(self) -> None:
        if not self.ready_to_record:
            return
        self.baseline_power_spectrum = record_power_spectrum(self.sdr_class, self.gain)

    def record_normalised_sky_power_spectrum(self) -> tuple[NDArray, NDArray]:
        if not self.ready_to_observe:
            return
        freq, pwr = record_power_spectrum(self.sdr_class, self.gain)
        __, baseline_pwr = self.baseline_power_spectrum
        pwr = pwr / baseline_pwr
        pwr = (pwr - pwr.min()) / pwr.ptp()
        return freq, pwr
