from __future__ import annotations

import argparse
import os
import tkinter as tk
import tkinter.font
import traceback
from pathlib import Path
from typing import Any, Callable, Optional, Union

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from numpy.typing import NDArray
from PIL import Image, ImageTk

from skao_trt.plotting import plot_power_spectrum_on_axes

from .obs_manager import ObsManager

PathLike = Union[os.PathLike, str]

ConfigDict = dict[str, Any]


def make_config(width_pixels: int) -> ConfigDict:
    REF_WIDTH = 1000
    REF_DPI = 80
    dpi = width_pixels / REF_WIDTH * REF_DPI
    return {
        "window.width": width_pixels,
        "window.fontsize": 10,
        "figure.aspect_ratio": 0.5,
        "figure.dpi": dpi,
        "figure.text.fontsize": 28,
        "figure.axislabels.fontsize": 14,
        "figure.ticklabels.fontsize": 12,
        "figure.title.fontsize": 18,
        "figure.text.x": 0.12,
        "figure.text.y": 0.70,
        "figure.text.dy": 0.1,
    }


def get_logo_path() -> Path:
    return Path(os.path.dirname(__file__)) / ".." / "img" / "skao_logo_bar.jpg"


def get_default_font_family() -> str:
    font = tkinter.font.nametofont("TkCaptionFont")
    return font.actual()["family"]


class LogoBar:
    """
    Wrapper for the Label containing the SKAO logo.
    """

    def __init__(self, parent: tk.Widget, imgpath: PathLike, *, width: int) -> None:
        image = Image.open(imgpath)
        aspect_ratio = image.width / image.height
        height = round(width / aspect_ratio)
        image = image.resize((width, height))
        image = ImageTk.PhotoImage(image)
        self.label = tk.Label(parent, image=image)
        # Must keep a reference to the image, see:
        # https://stackoverflow.com/a/16424553
        self.label.image = image

    def dispose(self) -> None:
        self.label.pack(side=tk.TOP, fill=tk.BOTH)


class InteractiveFigure:
    """
    Wrapper for the matplotlib Figure containing the power spectrum plot.
    """

    def __init__(self, parent: tk.Widget, config: ConfigDict) -> None:
        # Frame containing both the figure and the toolbar
        self.config = config

        width = config["window.width"]
        self.frame = tk.Frame(parent, width=width)

        height = round(config["figure.aspect_ratio"] * width)
        dpi = config["figure.dpi"]
        self.fig = Figure(figsize=(width / dpi, height / dpi), dpi=dpi)
        self.ax = self.fig.add_subplot(111)
        # Make sure the figure occupies the whole frame
        self.fig.subplots_adjust(left=0.07, right=0.97, bottom=0.1, top=0.95)
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.frame)
        self.clear_axes()
        self.navbar = NavigationToolbar2Tk(self.canvas, window=self.frame)

    def clear_axes(self) -> None:
        self.ax.clear()
        self.ax.axis("off")
        self.canvas.draw()

    def display_text(self, title: str, lines: list[str]):
        """
        Display given lines of text on the plotting area. Title is shown at
        the top, in bold and in a different color.
        """
        self.clear_axes()
        SKAO_BLUE = "#070068"
        SKAO_MAGENTA = "#E50869"

        fontsize = self.config["figure.text.fontsize"]
        x = self.config["figure.text.x"]
        y = self.config["figure.text.y"]
        ystep = self.config["figure.text.dy"]

        self.ax.text(
            x, y, title, fontsize=fontsize, fontweight="bold", color=SKAO_MAGENTA
        )
        y -= ystep

        for line in lines:
            self.ax.text(x, y, line, fontsize=fontsize, color=SKAO_BLUE)
            y -= ystep

        self.canvas.draw()

    def dispose(self) -> None:
        self.frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.navbar.pack(side=tk.BOTTOM, fill=tk.X, expand=True)

    def plot_power_spectrum(self, freq: NDArray, pwr: NDArray) -> None:
        self.clear_axes()
        plot_power_spectrum_on_axes(self.ax, freq, pwr)
        self.ax.set_xlabel(
            self.ax.get_xlabel(), fontsize=self.config["figure.axislabels.fontsize"]
        )
        self.ax.set_ylabel(
            self.ax.get_ylabel(), fontsize=self.config["figure.axislabels.fontsize"]
        )
        self.ax.set_ylim(0.0, 1.01 * pwr.max())
        self.ax.tick_params(
            axis="both",
            which="major",
            labelsize=self.config["figure.ticklabels.fontsize"],
        )
        self.ax.axis("on")  # don't forget to turn axes back on
        self.ax.set_title(
            "Hydrogen Line (HI) signal of the Milky Way",
            fontsize=self.config["figure.title.fontsize"],
        )
        self.canvas.draw()


class ButtonsBar:
    """
    Wrapper for a Frame containing buttons laid out horizontally.
    """

    def __init__(
        self,
        parent: tk.Widget,
        button_definitions: list[tuple[str, Callable]],
        *,
        font: tuple[str, int],
    ) -> None:
        self.frame = tk.Frame(parent)
        self.buttons: list[tk.Button] = []
        self.font = font
        for text, command in button_definitions:
            self._create_button(text, command=command)

    def _create_button(
        self, text: str, command: Optional[Callable] = None
    ) -> tk.Button:
        button = tk.Button(self.frame, text=text, command=command, font=self.font)
        self.buttons.append(button)

    def dispose(self) -> None:
        for col in range(len(self.buttons)):
            self.frame.grid_columnconfigure(col, weight=1)
        for i, button in enumerate(self.buttons):
            button.grid(column=i, row=0, sticky="ew")
        self.frame.pack(side=tk.BOTTOM, fill=tk.X)


class StatusBar:
    """
    Wrapper for the status bar Label.
    """

    def __init__(self, parent: tk.Widget, font: tuple[str, int]) -> None:
        self.label = tk.Label(
            parent,
            font=font,
            relief=tk.SUNKEN,
            anchor=tk.W,
            height=2,
        )
        self.update_status("Ready")

    def dispose(self) -> None:
        self.label.pack(side=tk.BOTTOM, fill=tk.X)

    def update_status(self, message: str) -> None:
        self.label.config(text="Status: " + message)


class TabletopApp:
    """
    Main application class.
    """

    def __init__(self, width_pixels: int = 1200, *, mock_device: bool = False) -> None:
        self.root = tk.Tk()
        self.root.title("SKAO Table-top Radio Telescope (TRT) App")
        self.config = make_config(width_pixels)

        self.logo_bar = None
        self.interactive_figure = None
        self.buttons_bar = None
        self.status_bar = None

        self.obs_manager = ObsManager(mock_device)
        self._create_widgets()
        self._dispose_widgets()
        self._display_welcome_text()

    def _create_widgets(self) -> None:
        font = (get_default_font_family(), self.config["window.fontsize"])
        config = self.config
        self.logo_bar = LogoBar(
            self.root, get_logo_path(), width=config["window.width"]
        )
        self.interactive_figure = InteractiveFigure(self.root, self.config)
        button_definitions = [
            ("Prepare", self.prepare),
            ("Observe", self.observe),
            ("Exit", self.exit),
        ]
        self.buttons_bar = ButtonsBar(self.root, button_definitions, font=font)
        self.status_bar = StatusBar(self.root, font=font)

    def _dispose_widgets(self) -> None:
        """
        Dispose the widgets in the main window once they are created.
        """
        self.logo_bar.dispose()
        # NOTE: dispose the widgets fom bottom to top
        self.status_bar.dispose()
        self.buttons_bar.dispose()
        self.interactive_figure.dispose()
        self.root.resizable(False, False)

    def _display_welcome_text(self) -> None:
        self.interactive_figure.display_text(
            "SKAO Table-top Radio Telescope (TRT) App",
            [
                "Short instruction manual:",
                "1. Prepare",
                "2. Observe as many times as you want",
            ],
        )

    def update_status(self, message: str) -> None:
        self.status_bar.update_status(message)
        # "Goofy way" to make sure the status bar updates
        # https://stackoverflow.com/a/16700254
        self.root.update_idletasks()

    def display_error(self, err: Exception) -> None:
        self.interactive_figure.display_text(
            "Something went wrong",
            [
                "Check status bar and terminal for more information.",
            ],
        )
        print(traceback.format_exc())
        self.update_status(f"Error: {err!s}")

    def prepare(self) -> None:
        self.update_status("Preparing ...")
        self.interactive_figure.display_text(
            "Preparation in progress",
            [
                "This step calibrates your RF device",
                "Searching for optimal gain ...",
            ],
        )

        # Find optimal gain
        try:
            self.obs_manager.find_optimal_gain(
                callback=lambda gain: self.update_status(f"Trying gain = {gain:.1f}")
            )
        except Exception as err:
            self.display_error(err)
            return

        self.interactive_figure.display_text(
            "Preparation in progress",
            [
                "This step calibrates your RF device",
                "Searching for optimal gain ... DONE.",
                "Recording baseline power spectrum ...",
            ],
        )

        # Record baseline power spectrum
        try:
            self.update_status("Recording baseline power spectrum ...")
            self.obs_manager.record_baseline_power_spectrum()
        except Exception as err:
            self.display_error(err)
            return

        self.update_status("Preparation complete!")
        optimal_gain = self.obs_manager.gain
        self.interactive_figure.display_text(
            "Preparation complete",
            [
                f"Optimum Gain = {optimal_gain:.2f}",
                "You can observe the Milky Way!",
                " - Remove RF termination",
                " - Connect your antenna to the Sawbird HI",
                " - Press Observe",
            ],
        )

    def observe(self) -> None:
        if not self.obs_manager.ready_to_observe:
            self.update_status("Please run the preparation step first!")
            return

        self.interactive_figure.clear_axes()
        self.interactive_figure.display_text(
            "Sky observation in progress",
            [
                "This should take about 30 seconds ...",
                "An interactive graph will appear at the end!",
            ],
        )
        self.update_status("Observing ...")

        try:
            freq, pwr = self.obs_manager.record_normalised_sky_power_spectrum()
        except Exception as err:
            self.display_error(err)
            return

        self.update_status("Plottting ...")
        self.interactive_figure.plot_power_spectrum(freq, pwr)
        self.update_status("Observation finished!")

    def exit(self) -> None:
        self.root.destroy()

    def run(self) -> None:
        self.root.mainloop()


def parse_args():
    parser = argparse.ArgumentParser(
        description="Launch the SKAO Table-top Radio Telescope (TRT) App"
    )
    parser.add_argument(
        "--mock",
        action="store_true",
        help="Use a mock version of the SDR device which generates white noise",
    )
    parser.add_argument(
        "--width",
        type=int,
        default=1200,
        help="Main window width in pixels",
    )
    return parser.parse_args()


def main():
    args = parse_args()
    TabletopApp(args.width, mock_device=args.mock).run()


if __name__ == "__main__":
    main()
